#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "keymap.h"
#include "translate.h"
#include <linux/vt.h>
#include <linux/input.h>
#include <getopt.h>

#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define OFF(x)  ((x)%BITS_PER_LONG)
#define BIT(x)  (1UL<<OFF(x))
#define LONG(x) ((x)/BITS_PER_LONG)
#define test_bit(bit, array)    ((array[LONG(bit)] >> OFF(bit)) & 1)

#define ESC 27
#define INDEFAULT "/dev/input/event2"

int shifts[16]={0,0,0,0,0,0,0,0,0,0,0,0};
int keys_down[128];
int tty, indev;
int ttynum;
char ttyfilename[24];
static int npadch = -1;			/* -1 or number assembled on pad */
int console_control;

int find_keyboard() {
    int i;
    for(i=0;i<16;i++) {
        int fd;
        char s[64];
        sprintf(s,"/dev/input/event%d",i);
        fd=open(s,O_RDONLY);
        if(fd>=0) {
            unsigned short id[4];
            unsigned long bit[EV_MAX][NBITS(KEY_MAX)];
            ioctl(fd, EVIOCGID, id);
	    memset(bit, 0, sizeof(bit));
            ioctl(fd, EVIOCGBIT(0, EV_MAX*3), bit[0]);
            if(test_bit(1,bit[0])) {
                ioctl(fd, EVIOCGBIT(1, EV_MAX*3), bit[1]);
                
                if(test_bit(1,bit[1]) && test_bit(17,bit[1])) {
                    close(fd);
                    return i;
                }
            }
            close(fd);
        }
    }
    return -1;
}

int is_a_console(int fd) 
{
  char arg;
  
  arg = 0;
  return (ioctl(fd, KDGKBTYPE, &arg) == 0
	  && ((arg == KB_101) || (arg == KB_84)));
}

static int open_a_console(char *fnam) 
{
  int fd;
  
  fd = open(fnam, O_RDONLY);
  
  /* if failed, fail */
  if (fd < 0)
      return -1;
  
  /* if not a console, fail */
  if (! is_a_console(fd))
    {
      close(fd);
      return -1;
    }
  
  /* success */
  return fd;
}

int get_console_fd(void) 
{
  int fd;

  fd = open_a_console("/dev/tty");
  if (fd >= 0)
    return fd;
  
  fd = open_a_console("/dev/tty0");
  if (fd >= 0)
    return fd;
  
  fd = open_a_console("/dev/console");
  if (fd >= 0)
    return fd;
  
  for (fd = 0; fd < 3; fd++)
    if (is_a_console(fd))
      return fd;
  
#if 0
  fprintf(stderr, 
          "Couldnt get a file descriptor referring to the console\n");
#endif
  return -1;		/* total failure */
}

void send(int value)
{
    if(ioctl(tty,TIOCSTI,&value)) {
       close(tty);
       tty=open(ttyfilename,O_RDONLY);
       ioctl(tty,TIOCSTI,&value);
    }
};

int raw_handle_key(int key)
{
   send(key);
}

int do_console(c) {
   int fd;
   int newtty;
   struct vt_stat vts;
   
   if((ttynum==c)||!console_control)return 0;

   ttynum=c;
   snprintf(ttyfilename,22,"/dev/tty%d",ttynum);
    
   newtty=open(ttyfilename,O_RDONLY);

   if((newtty!=-1)){
      close(tty);
      tty=newtty;
      if(console_control==3)return 0;
      fd=get_console_fd();
      ioctl(fd,VT_GETSTATE,&vts);
      ioctl(fd,VT_ACTIVATE,c);
      ioctl(fd,VT_WAITACTIVE,c);
      if((c>12)&&(c<25)&&(vts.v_active<13)&&(console_control==2)) {
         ioctl(fd,VT_ACTIVATE,vts.v_active);
         ioctl(fd,VT_WAITACTIVE,vts.v_active);
      }
      if(fd>3)close(fd);
   }
}

do_cursor(int c) {
   static const char *cur_chars = "BDCA";
   send(ESC);
   send('[');
   send(cur_chars[c]); 
}

do_fn(int c) {
   char *p;

   p=func_table[c];
   if(p) do {send(*p); p++; } while(*p);
}

int cooked_handle_key(int key) {
   int i;
   int value,v;
   int leds, alt_ch;
   int rep;
   char *p, output[2];
   static const char *pad_chars = "0123456789+-*/\015,.?()";
   static ushort *map;
   int kmp,j;

/* first handle shifts */
   leds=0;
   alt_ch=0;
   rep=0;
   if(key<128) {
      if(keys_down[key])rep=1;
      keys_down[key]=1; 
   } else keys_down[key-128]=0;
         
   kmp=0;
   for(j=0;j<4;j++)kmp+=shifts[j]?1<<j:0;
   kmp ^=shifts[12]?2:0;
   map=key_maps[kmp];
    
   if(!map)return -1;
    
   value=map[key];
   v=value&0xff;
   if(key<0x80) { 
      switch((value>>8)&0x0f){
           case 0x0b:
           case 0x00:
              if (((v&0xdf)>='A')&&((v&0xdf)<='Z')&&shifts[14])
                  v^=32;
              if(v)send(v);
              break;
           case 0x01:
              do_fn(v);
              break;
           case 0x02:
              switch(v) {
                 case 0x01: send(0x0d); break;
                 case 0x07: if(!rep) {
                               shifts[14]=1-shifts[14]; 
                               leds=1; 
                            }
                      break;
                 case 0x08: if(!rep) {
                               shifts[13]=1-shifts[13]; 
                               leds=1; 
                            }
                      break;
              }
              break;
           case 0x03: /* numeric pad */
              if(shifts[13]){
                 send(pad_chars[v]);
              } else {
                 switch(v){
                    case 2: do_cursor(0); break; 
                    case 4: do_cursor(1); break; 
                    case 6: do_cursor(2); break; 
                    case 8: do_cursor(3); break; 
                    case 0: do_fn(21); break;
                    case 1: do_fn(23); break;
                    case 3: do_fn(25); break;
                    case 5: send(ESC); send('['); send('G'); break;
                    case 7: do_fn(20); break;
                    case 9: do_fn(24); break;
                    case 16: do_fn(22); break;
                    default: send(pad_chars[v]);
                 }
              }
              break;
           case 0x04: 
                    
              break;
           case 0x05: 
              do_console(v+1);      
              break;
           case 0x06:
              do_cursor(v);
              break;
           case 0x07: if(!rep)
              shifts[v]+=1;       
              break;
           case 0x08: 
                    
              break;
           case 0x09: { /* ascii */ 
 	      int base;
                
	      if (v < 10)    /* decimal input of code, while Alt depressed */
                 base = 10;
              else {       /* hexadecimal input of code, while AltGr depressed */
	         v -= 10; base = 16;
              }
	      if (npadch == -1) npadch = value; else npadch = npadch * base + value;
              }
              break;
           case 0x0a: 
              switch(v) {
                 case 0x01: if(!rep) {
                               shifts[12]=1-shifts[12]; 
                               leds=1; 
                            }
                      break;
              }
              break;
           case 0x0c: 
                   
              break;
           case 0x0d: 
                 
              break;
      }
   } else {
      switch((value>>8)&0x0f){
           case 0x07:
              if(shifts[v])shifts[v]-=1;       
              break;
      }
   }
   
   if(leds){
      struct input_event ev;
      int i,j;
      
      for(j=0;j<3;j++) {
          ev.type=EV_LED;
          ev.code=j;
          ev.value=shifts[12 + ((1+j)%3)] & 1;
          i=write(indev, &ev, sizeof(struct input_event));
      }
   }
   
   if((alt_ch)&&(npadch!=-1)) {
      send(npadch);
      npadch=-1;
   }
   
}

int main (int argc, char *argv[])
{
    unsigned char ch,c2,c3,c4;
    struct input_event ev;
    int brk=0;
    int raw,leds;
    int kbmode;
    int i;
    char output[2], indevname[130];
    int devnum=-1;
//    if(fork()!=0) exit(0);
    daemon(0,0);

    strcpy(indevname, INDEFAULT);
    
    console_control=2;
 
    i=1;
    while(i) {
        int c;
        c=getopt(argc, argv, "cd:fkns");
        switch(c) {
            case 'c':
                console_control=0;
                break;
            case 'n':
                console_control=1;
                break;
            case 's':
                console_control=2;
                break;
            case 'k':
                console_control=3;
                break;
            case 'd':
                strncpy(indevname, optarg, 128);
                break;
            case 'f':
                devnum=find_keyboard();
                if(devnum>-1)sprintf(indevname, "/dev/input/event%d", devnum);
                break;
            case -1:
                i=0;
                break;
#if 0
            default:
                printf("Usage: usb2key [options] <tty_num>\n\n"
                "options:\n"
                "\t-d dev\tselect input device\n"
                "\t-n\tnormal console switching\n"
                "\t-c\tno console switching\n"
                "\t-k\tno switching, but changing the console that recieves\n"
                "\t\tkeypresses is possible\n\n"
                );
                exit(1);
                break;
#endif
        }
    
    }
    
    ttynum=strtol(argv[optind],NULL,0);
    snprintf(ttyfilename,22,"/dev/tty%d",ttynum);
    indev=open(indevname,O_RDWR);
    tty=open(ttyfilename,O_RDONLY);

    {
      struct input_event ev;
      int i,j;
      
      for(j=0;j<3;j++) {
          ev.type=EV_LED;
          ev.code=j;
          ev.value=0;
          i=write(indev, &ev, sizeof(struct input_event));
      }
    }
    
    for(i=0;i<128;i++)keys_down[i]=0;

    do {
       int i;
       
       i=read(indev, &ev, sizeof(struct input_event));
       if(i<=0) {
           if(devnum==-1)exit(1);
           close(indev);
           sleep(5);
           devnum=find_keyboard();
           if(devnum==-1)exit(2);
           sprintf(indevname, "/dev/input/event%d", devnum);
           indev=open(indevname,O_RDWR);
           if(indev==-1)exit(3);
           ev.type=-1;
       }
       if(ev.type == EV_KEY) {
           ioctl(tty,KDGKBMODE,&kbmode);
           if((kbmode&1)==0)raw=1; else raw=0;
    
           ch=ev.code;
           if(ch<128) {
               brk = (ev.value==0) ? 0x80 : 0 ;
               if(raw) {
                   send(brk|ch);
               } else {
                   cooked_handle_key(brk|ch);
               }
           }
       }
    }  while(1);
}
